/**
 * Units:
 * - Point : 1 unit of signal data as sent by the simulator. This is a decimal
 * representation. The width of the screen for instance if expressed in points,
 * which is not an integer
 *
 * - Vertice: 1 unit of signal data in the javascript buffer. This is an integer
 *
 * - Coordonate:  1 unit of signal data in the vertex buffer.
 *   Each vertex is represented by three floating-point numbers that correspond to the x, y, and z coordinates
 *
 * - Clipspace: unit of screen in webgl canvas. 0-2
 *
 */

import {ScreenState} from 'src/store/state';

export function verticeSizeToCoordonateSize(size: number): number {
  return size * 3;
}

export function coordonateSizeToVerticeSize(size: number): number {
  return size / 3;
}

// A point is represented by 3 coordonates (x,y,z)
// A point is a decimal value
// A coordonate is the nearest integer multiple of 3
export function pointSizeToCoordonateSize(size: number): number {
  const decimalValue = 3 * size;
  return 3 * Math.floor(Math.abs(decimalValue) / 3);
}

export function pointSizeToPixelSize(
  screenState: ScreenState,
  pointSize: number
): number {
  const pointsPerWindow = screenSizeToPointSize(screenState);

  const ratio = screenState.containerWidth / pointsPerWindow;
  return pointSize * ratio;
}

export function pixelSizeToPointSize(
  screenState: ScreenState,
  pixelSize: number
): number {
  const pointsPerWindow = screenSizeToPointSize(screenState);

  const ratio = pointsPerWindow / screenState.containerWidth;
  return pixelSize * ratio;
}

export function pixelSizeToClipspaceSize(
  screenState: ScreenState,
  pixelSize: number
): number {
  const pixelToClipspace = 2 / screenState.containerWidth;
  return pixelSize * pixelToClipspace;
}

export function pixelSizeToVerticeSize(
  screenState: ScreenState,
  pixelSize: number
): number {
  const nbrVertices =
    -(pixelSize * screenSizeToPointSize(screenState)) /
    screenState.containerWidth;
  return Math.round(nbrVertices);
}

export const screenSizeToPointSize = ({
  mmToPx,
  containerWidth,
  displayRate,
  pitch,
}: ScreenState): number => {
  const windowWidthInMm = mmToPx * containerWidth;
  const pointsPerSecond = windowWidthInMm / displayRate;

  return (pointsPerSecond * 1000) / pitch;
};
